Require Import Arith List Program FunctionalExtensionality Heq.
Require Import util.
Require Import ssreflect ssrbool Heq.
Require Import cesd sysf sysfopsem logrel.

Set Implicit Arguments.

Module EgCtxManipulate (ObsLeMod: ObsLeSig).

Module ObsLeModInstance := ObsLeMod.
Module LogRelMod := LogicalRelationFunctor (ObsLeModInstance).
Export LogRelMod.

(*
   PushPop relation is a bisimulation.
*)

Definition pushpop_related_dump (d1 d2: Dump) :=
  d1 = d2 \/
  exists d', exists d'', exists c', exists c'', exists e, exists s,
    d1 = d'++(c'++[PushN 0;Pop]++c'',e,s)::d'' /\
    d2 = d'++(c'++c'',e,s)::d''.

Definition pushpop_related (cesd1 cesd2: CESD) : Prop :=
  let (ces1,d1):=cesd1 in let (ce1,s1):=ces1 in let (c1,e1):=ce1 in
  let (ces2,d2):=cesd2 in let (ce2,s2):=ces2 in let (c2,e2):=ce2 in
    pushpop_related_dump ((c1,e1,s1)::d1) ((c2,e2,s2)::d2).

Lemma pushpop_simulation1: forall cesd1 cesd2 cesd1',
  pushpop_related cesd1 cesd2 -> step cesd1 cesd1' ->
  step cesd1' cesd2 \/
  exists cesd2', step cesd2 cesd2' /\ pushpop_related cesd1' cesd2'.
Proof.
move => [[[c1 e1] s1] d1] [[[c2 e2] s2] d2] [[[c1' e1'] s1'] d1'] Hrel Hstep.
destruct Hrel; first by depdestr H; right; eexists; split;[Apply | left].
destruct H as [d' [d'' [c' [c'' [e [s [Hrel1 Hrel2]]]]]]].
destruct d'.
- depdestr Hrel1, Hrel2.
  depdestr Hstep; 
  try (by depdestr c'; try (by left; cesd_step); right; eexists; simpl; split; 
   [cesd_step 
   |simpl; solve 
    [by right; eexists [];eexists;eexists;eexists;eexists;eexists; split; simpl;
        try rewrite <- (app_ass c'0)
    |by right; eexists [_];eexists;eexists;eexists;eexists;eexists; split; simpl
    |by left; simpl]]).
- depdestr Hrel1; right.
  depdestr Hstep;
  try (by eexists; simpl; split;
   [cesd_step
   |simpl; right; eexists;eexists;eexists;eexists;eexists;eexists; split; simpl;
    solve [by rewrite -> (app_comm_cons (A:=Code*MEnv*Stack))
          |by do 2 rewrite -> (app_comm_cons (A:=Code*MEnv*Stack))]]).
  - by depdestr d'; eexists; simpl; (split; first by cesd_step); simpl;
     right; [eexists[]|eexists];eexists;eexists;eexists;eexists;eexists; simpl;
      [split | split;rewrite -> (app_comm_cons (A:=Code*MEnv*Stack))].
  - by depdestr d'; eexists; simpl; (split; first by cesd_step); simpl;
     right; [eexists[]|eexists];eexists;eexists;eexists;eexists;eexists; simpl;
      [split; (try rewrite app_ass; try do 2 rewrite <- app_comm_cons; auto)
      |split; rewrite -> (app_comm_cons (A:=Code*MEnv*Stack))].
Qed.

Lemma pushpop_simulation2: forall cesd1 cesd2 cesd2',
  pushpop_related cesd1 cesd2 -> step cesd2 cesd2' ->
  exists k, exists cesd1', kstepreduce (S k) cesd1 cesd1' /\ 
    pushpop_related cesd1' cesd2'.
Proof.
move => [[[c1 e1] s1] d1] [[[c2 e2] s2] d2] [[[c2' e2'] s2'] d2'] Hrel Hstep.
destruct Hrel; first 
 by depdestr H; exists 0; eexists; split; [econstructor;[Apply| auto] | left].
destruct H as [d' [d'' [c' [c'' [e [s [Hrel1 Hrel2]]]]]]].
destruct d'.
- depdestr Hrel1, Hrel2.
  depdestr Hstep; 
   try (by depdestr c'; eexists; eexists; [ split; [by ksteps | left] |
     split; [by ksteps | solve [by left |by right; solve 
      [by eexists[];eexists;eexists;eexists;eexists;eexists; simpl; split;
          try rewrite <- (app_ass c'0)
      |by eexists[_];eexists;eexists;eexists;eexists;eexists; simpl; split
      |by eexists;eexists;eexists;eexists;eexists;eexists; simpl; split;
       try rewrite -> (app_comm_cons (A:=Code*MEnv*Stack))]]]]).
- depdestr Hrel1.
  depdestr Hstep; try (by
   eexists; eexists; 
     split; [by ksteps | by right; 
      eexists;eexists;eexists;eexists;eexists;eexists; simpl; split; solve
      [by rewrite -> (app_comm_cons (A:=Code*MEnv*Stack))
      |by do 2 rewrite -> (app_comm_cons (A:=Code*MEnv*Stack))]]).
  - by depdestr d'; eexists; eexists; simpl; (split; first by ksteps); simpl;
     right; [eexists[]|eexists];eexists;eexists;eexists;eexists;eexists; simpl;
      [split | split;rewrite -> (app_comm_cons (A:=Code*MEnv*Stack))].
  - by depdestr d'; eexists; eexists; simpl; (split; first by ksteps); simpl;
     right; [eexists[]|eexists];eexists;eexists;eexists;eexists;eexists; simpl;
      [split; (try rewrite app_ass; try do 2 rewrite <- app_comm_cons; auto)
      |split; rewrite -> (app_comm_cons (A:=Code*MEnv*Stack))].
Qed.

Lemma pushpop_relation_steps: forall k cesd1 cesd2,
  pushpop_related cesd1 cesd2 -> stepsfor k cesd2 -> stepsfor k cesd1.
Proof.
move => gk; generalize_induction gk.
induction gk; first by intros; inversion H; eexists; eexists; auto.
move => k Hk cesd1 cesd2 Hrel Hred; destruct k; first by eexists; eexists; auto. 
destruct (stepsfor_inv Hred) as [cesd2' [Hstep Hsteps]]; clear Hred.
destruct (pushpop_simulation2 _ Hrel Hstep) as [j [cesd1' [Hred Hrel']]].
depdestr Hred.
eapply stepsfor_step_strict; first by Apply.
simpl; eapply stepsfor_kstepreduce_rev; first by Apply.
eapply IHgk; [omega | Apply | apply stepsfor_decr with k; [omega | Apply]].
Qed.

Lemma pushpop_relation_terminate: forall cesd1 cesd2,
  pushpop_related cesd1 cesd2 -> terminate cesd2 -> terminate cesd1.
Proof.
move => cesd1 cesd2 Hrel Hterm Hdiv; apply Hterm; clear Hterm.
move => k; specialize (Hdiv (S (S k))); move: cesd1 cesd2 Hrel Hdiv. 
induction k; first by intros; eexists; eexists; auto.
move => cesd1 cesd2 Hrel Hred.
destruct (stepsfor_inv Hred) as [cesd2' [Hstep Hsteps]]; clear Hred.
destruct (pushpop_simulation1 _ Hrel Hstep) 
 as [ |[cesd2'' [Hstep2 Hrel2]]].
- destruct (stepsfor_inv Hsteps) as [cesd2a [Hstepa Hstepsa]].
  erewrite step_uniq; Apply.
- eapply stepsfor_step_strict; first by Apply.
  eapply IHk; Apply.
Qed.

(*
   Context Manipulation Example
*)

Definition ctxman n :=
  [ PushN n; PushN 1; Sel [Join; PushN 0; Pop] [] ].

Theorem ctxman_n_realizes_n : forall n, 
  #c (ctxman n) ==cc== ^@VINT 0 [] n.
Proof.
move => n; split.
- eexists; split; first by apply ObsLeC_refl.
  move => k srel Hsrel k0 e te Hk0k Hete j c0 e0 s0 d0 Hcont0 k1 Hk1k0 Hk1j.
  unfold ctxman; ssteps.
  eapply pushpop_relation_steps with (c0,e++e0,VN n::s0,d0).
  - by right; eexists[];eexists;eexists;eexists[];eexists;eexists; simpl; split;
       [ | rewrite <- app_nil_end].
  eapply Hcont0.
  - by eexists; split; [unfold_SubVE; eexists | simpl].
  - by Apply. - by [].
- eexists; split; first by apply ObsLeC_refl.
  move => p srel e te Hte tv Htv c0 e0 s0 d0 Hcont0.
  move :Htv; unfold_SubVE; move => Htv; clear p.
  destruct Htv as [k Htv]; inversion Htv; Hsubst; clear Htv.
  tsteps; apply pushpop_relation_terminate with (c0,e++e0,VN n::s0,d0).
  - by right; eexists[];eexists;eexists;eexists[];eexists;eexists; simpl; split;
       [ | rewrite <- app_nil_end].
  - by apply Hcont0.
Qed.

End EgCtxManipulate.